﻿using System.Linq;
using System.Reflection;
using LinqToLdap.Mapping;
using LinqToLdap.Tests.ClassMapAssembly;
using Moq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SharpTestsEx;
using System;

namespace LinqToLdap.Tests.Mapping
{
    [TestClass]
    public class DirectoryMapperTest
    {
        private Mock<IClassMap> _classMap;
        private DirectoryMapper _mapper;

        public string Property1 { get; set; }


        [TestInitialize]
        public void SetUp()
        {
            _classMap = new Mock<IClassMap>();
            _mapper = new DirectoryMapper();
            //_mapper.AutoMapWith(type =>
            //                        {
            //                            var mapType = typeof (AutoClassMap<>).MakeGenericType(type);

            //                            return
            //                                (IClassMap) Activator.CreateInstance(mapType);
            //                        });

            //_mapper.AttributeMapWith(type =>
            //                             {
            //                                 var mapType = typeof (AttributeClassMap<>).MakeGenericType(type);

            //                                 return
            //                                     (IClassMap) Activator.CreateInstance(mapType);
            //                             });
        }

        [TestMethod]
        public void Map_NonMappedStandardClass_CachesObjectMapping()
        {
            //prepare
            var objectMapping = new Mock<IObjectMapping>();
            objectMapping.Setup(m => m.Type)
                .Returns(typeof(DirectoryMapperTest));
            _classMap.Setup(m => m.ToObjectMapping())
                .Returns(objectMapping.Object);
            _classMap.Setup(m => m.PerformMapping("nc", "oc", false, new[] { "oc" }, false))
                .Returns(_classMap.Object);
            _classMap.Setup(m => m.Type)
                .Returns(typeof(DirectoryMapperTest));

            //act
            _mapper.Map(_classMap.Object, "nc", new[] { "oc" }, false, "oc", false);

            //assert
            _mapper.GetMappings().Should().Have.Count.EqualTo(1);
            _mapper.GetMappings()[GetType()].Should().Be.EqualTo(objectMapping.Object);
            _classMap.Verify(c => c.PerformMapping("nc", "oc", false, new[] { "oc" }, false));
            _classMap.Verify(c => c.Validate());
        }

        [TestMethod]
        public void Map_NonPremappedType_CreatesAutoClassMap()
        {
            //act
            var mapping = _mapper.Map<DirectoryMapperTest>("context");

            //assert
            _mapper.GetMappings().Should().Have.Count.EqualTo(1);
            mapping.Type.Should().Be.EqualTo(GetType());
        }

        [TestMethod]
        public void Map_NonPremappedTypeAndCustomAutoClassMapper_CareatesAutoClassMap()
        {
            //prepare

            var objectMapping = new Mock<IObjectMapping>();
            objectMapping.Setup(m => m.Type)
                .Returns(typeof(DirectoryMapperTest));
            _classMap.Setup(m => m.ToObjectMapping())
                .Returns(objectMapping.Object);
            _classMap.Setup(m => m.PerformMapping("nc", "oc", true, new[] { "oc" }, true))
                .Returns(_classMap.Object);
            _classMap.Setup(m => m.Type)
                .Returns(typeof(DirectoryMapperTest));
            _mapper.AutoMapWith(t => _classMap.Object);

            //act
            var mapping = _mapper.Map<DirectoryMapperTest>("nc", objectCategory: "oc", objectClasses: new[] { "oc" });

            //assert
            _mapper.GetMappings().Should().Have.Count.EqualTo(1);
            mapping.Type.Should().Be.EqualTo(GetType());
            mapping.Should().Be.SameInstanceAs(objectMapping.Object);
            _classMap.Verify(c => c.PerformMapping("nc", "oc", true, new[] { "oc" }, true));
            _classMap.Verify(c => c.Validate());
        }

        [TestMethod]
        public void Map_NonPremappedTypeAndCustomAttributeClassMapper_CareatesAutoClassMap()
        {
            //prepare
            var objectMapping = new Mock<IObjectMapping>();
            objectMapping.Setup(m => m.Type)
                .Returns(typeof(AttributeClass));
            _classMap.Setup(m => m.ToObjectMapping())
                .Returns(objectMapping.Object);
            _classMap.Setup(m => m.PerformMapping("nc", "oc", true, new[] { "oc" }, true))
                .Returns(_classMap.Object);
            _classMap.Setup(m => m.Type)
                .Returns(typeof(AttributeClass));
            _mapper.AttributeMapWith(t => _classMap.Object);

            //act
            var mapping = _mapper.Map<AttributeClass>("nc", objectCategory: "oc", objectClasses: new[] { "oc" });

            //assert
            _mapper.GetMappings().Should().Have.Count.EqualTo(1);
            mapping.Type.Should().Be.EqualTo(typeof(AttributeClass));
            mapping.Should().Be.SameInstanceAs(objectMapping.Object);
            _classMap.Verify(c => c.PerformMapping("nc", "oc", true, new[] { "oc" }, true));
            _classMap.Verify(c => c.Validate());
        }

        [TestMethod]
        public void Map_ObjectClassAndObjectClasses_ThrowsArgumentException()
        {
            //act
            Executing.This(() => _mapper.Map<DirectoryMapperTest>("context", objectClass: "oc", objectClasses: new[] { "oc" }))
                .Should().Throw<ArgumentException>().And.Exception.Message.Should().Be.EqualTo("objectClass and objectClasses cannot both have a value.");
        }

        [TestMethod]
        public void AddMappingFromAssembly_Assembly_AddsMappings()
        {
            //prepare
            var assembly = Assembly.GetAssembly(typeof(AssemblyTestClass));

            //act
            _mapper.AddMappingsFrom(assembly);

            //assert
            var mappings = _mapper.GetMappings();
            mappings.Should().Have.Count.EqualTo(3);
            mappings.ContainsKey(typeof(AssemblyTestClass)).Should().Be.True();
            mappings.ContainsKey(typeof(AssenblyTestClass2)).Should().Be.True();
            mappings.ContainsKey(typeof(AssemblyTestClassSub)).Should().Be.True();
        }

        [TestMethod]
        public void AddMappingFromAssembly_AssemblyNameWithoutDllEnd_AddsMappings()
        {
            //act
            _mapper.AddMappingsFrom("LinqToLdap.Tests.ClassMapAssembly");

            //assert
            var mappings = _mapper.GetMappings();
            mappings.Should().Have.Count.EqualTo(3);
            mappings.Keys.Select(k => k.ToString()).Should().Contain(typeof (AssemblyTestClass).ToString());
            mappings.Keys.Select(k => k.ToString()).Should().Contain(typeof(AssenblyTestClass2).ToString());
            mappings.Keys.Select(k => k.ToString()).Should().Contain(typeof(AssemblyTestClassSub).ToString());
        }

        [DirectorySchema("test")]
        private class AttributeClass
        {
            [DirectoryAttribute]
            public string Property1 { get; set; }
        }
    }
}
