﻿/*
 * LINQ to LDAP
 * http://linqtoldap.codeplex.com/
 * 
 * Copyright Alan Hatter (C) 2010-2014
 
 * 
 * This project is subject to licensing restrictions. Visit http://linqtoldap.codeplex.com/license for more information.
 */

using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using LinqToLdap.Exceptions;

namespace LinqToLdap.Mapping
{
    /// <summary>
    /// Uses attributes to map a class
    /// </summary>
    /// <typeparam name="T">The class to map</typeparam>
    public class AttributeClassMap<T> : ClassMap<T> where T : class 
    {
        /// <summary>
        /// Maps class information for <typeparamref name="T"/>.
        /// </summary>
        /// <exception cref="MappingException">Thrown if <typeparamref name="T"/> does not have a <see cref="DirectorySchemaAttribute"/>.</exception>
        public override IClassMap PerformMapping(string namingContext = null, string objectCategory = null, bool includeObjectCategory = true, IEnumerable<string> objectClasses = null, bool includeObjectClasses = true)
        {
            var type = typeof(T);
            var schemaAttribute = type
                .GetCustomAttributes(typeof(DirectorySchemaAttribute), true)
                .Cast<DirectorySchemaAttribute>()
                .FirstOrDefault();

            if (schemaAttribute == null)
            {
                throw new MappingException("DirectorySchemaAttribute not found for " + typeof(T).FullName);
            }

            NamingContext(namingContext.IsNullOrEmpty() ? schemaAttribute.NamingContext : namingContext);
            if (!objectCategory.IsNullOrEmpty())
            {
                ObjectCategory(objectCategory, includeObjectCategory);
            }
            else
            {
                ObjectCategory(schemaAttribute.ObjectCategory, schemaAttribute.IncludeObjectCategory);
            }
            
            if (objectClasses != null)
            {
                ObjectClasses(objectClasses, includeObjectClasses);
            }
            else
            {
                ObjectClasses(schemaAttribute.ObjectClasses, schemaAttribute.IncludeObjectClasses);
            }

            var properties = type.GetProperties(Flags)
                .Where(p => p.GetCustomAttributes(typeof(DirectoryAttributeAttribute), true).Any() &&
                            p.GetGetMethod(true) != null && p.GetSetMethod(true) != null)
                .Select(p =>
                        new KeyValuePair<DirectoryAttributeAttribute, PropertyInfo>(
                            p.GetCustomAttributes(typeof(DirectoryAttributeAttribute), true)
                                .Cast<DirectoryAttributeAttribute>()
                                .FirstOrDefault(),
                            p));

            properties.ToList()
                .ForEach(p =>
                {
                    var property = MapPropertyInfo(p.Value)
                        .Named(p.Key.AttributeName);

                    if (p.Key.EnumStoredAsInt)
                        property.EnumStoredAsInt();

                    if (p.Key.StoreGenerated)
                        property.StoreGenerated();

                    if (p.Key.ReadOnly)
                        property.ReadOnly();

                    property.DateTimeFormat(p.Key.DateTimeFormat);
                });

            var distinguishedName = type.GetProperties(Flags)
                .Where(p => p.GetCustomAttributes(typeof(DistinguishedNameAttribute), true).Any() &&
                                                     p.GetGetMethod(true) != null && p.GetSetMethod(true) != null)
#if NET35
                .Select(p => new LinqToLdap.Helpers.TwoTuple<PropertyInfo, DistinguishedNameAttribute>(p,
#else
                .Select(p => new System.Tuple<PropertyInfo, DistinguishedNameAttribute>(p,
#endif
                
                    p.GetCustomAttributes(typeof(DistinguishedNameAttribute), true).Cast<DistinguishedNameAttribute>().First()))
                .FirstOrDefault();

            if (distinguishedName != null)
            {
                DistinguishedName(distinguishedName.Item1, distinguishedName.Item2.AttributeName);
            }

            return this;
        }
    }
}
