﻿/*
 * LINQ to LDAP
 * http://linqtoldap.codeplex.com/
 * 
 * Copyright Alan Hatter (C) 2010-2014
 
 * 
 * This project is subject to licensing restrictions. Visit http://linqtoldap.codeplex.com/license for more information.
 */

using System.Collections.Generic;
using System.DirectoryServices.Protocols;
using LinqToLdap.Collections;
using LinqToLdap.Mapping;

namespace LinqToLdap.Transformers
{
    internal class ResultTransformer : IResultTransformer
    {
        protected readonly IDictionary<string, string> QueriedProperties;
        protected readonly IObjectMapping ObjectMapping;
        private readonly bool _setOriginalValues;

        public ResultTransformer(IDictionary<string, string> queriedProperties, IObjectMapping mapping, bool setOriginalValues = true)
        {
            QueriedProperties = queriedProperties;
            ObjectMapping = mapping;
            _setOriginalValues = setOriginalValues;
        }

        public virtual object Transform(SearchResultEntry entry)
        {
            object instance;
            if (!ObjectMapping.IsForAnonymousType)
            {
                instance = ObjectMapping.Create();
                bool isDirectoryObject = instance is IDirectoryObject;
                var originalValues = isDirectoryObject && _setOriginalValues
                                         ? new List<SerializableKeyValuePair<string, object>>(ObjectMapping.Properties.Count)
                                         : null;
                foreach (var queriedProperty in QueriedProperties)
                {
                    var property = ObjectMapping.GetPropertyMapping(queriedProperty.Key);

                    if (property.IsDistinguishedName)
                    {
                        var value = entry.DistinguishedName;
                        property.SetValue(instance, value);
                        if (originalValues != null)
                        {
                            originalValues.Add(new SerializableKeyValuePair<string, object>(property.PropertyName, value));
                        }
                    }
                    else
                    {
                        if (entry.Attributes.Contains(queriedProperty.Value))
                        {
                            property.SetValue(instance, GetValue(entry, property));
                            if (originalValues != null)
                            {
                                originalValues.Add(new SerializableKeyValuePair<string, object>(property.PropertyName, GetValue(entry, property)));
                            }
                        }
                        else
                        {
                            if (originalValues != null)
                            {
                                originalValues.Add(new SerializableKeyValuePair<string, object>(property.PropertyName, property.Default()));
                            }
                        }
                    }
                }
                if (originalValues != null)
                {
// ReSharper disable PossibleNullReferenceException
                    (instance as IDirectoryObject).OriginalValues = new OriginalValuesCollection(originalValues);
// ReSharper restore PossibleNullReferenceException
                }
            }
            else
            {
                var parameters = new object[ObjectMapping.Properties.Count];
                int count = 0;
                foreach (var property in ObjectMapping.GetPropertyMappings())
                {
                    object value;
                    if (property.IsDistinguishedName)
                    {
                        value = entry.DistinguishedName;
                    }
                    else
                    {
                        value = entry.Attributes.Contains(property.AttributeName)
                                    ? GetValue(entry, property)
                                    : property.Default();
                    }
                    
                    parameters[count++] = value;
                }
                instance = ObjectMapping.Create(parameters);
            }
            return instance;
        }

        public virtual object Default()
        {
            return null;
        }

        protected object GetValue(SearchResultEntry entry, IPropertyMapping property)
        {
            var value = entry.Attributes[property.AttributeName];

            return property.FormatValueFromDirectory(value, entry.DistinguishedName);
        }
    }
}
